import pickle

import gdb  # pylint: disable=E0401

from . import data_extractor
from . import util

try:
    import scipy.io
    SCIPY_AVAILABLE = True
except ImportError:
    SCIPY_AVAILABLE = False


class SaveMat(gdb.Command):
    def __init__(self):
        super(SaveMat, self).__init__("savemat", gdb.COMMAND_OBSCURE)

    def invoke(self, args, from_tty):
        if not SCIPY_AVAILABLE:
            raise RuntimeError("Scipy not available")

        out = {}
        filename, var = args.split()

        for v in var:
            data = data_extractor.extract_var(v)
            base_var, _ = util.split_variable_and_slice(v)
            dict_name = util.strip_non_alphanumeric(base_var)
            out[dict_name] = data

        scipy.io.savemat(filename, out)


class SavePy(gdb.Command):
    def __init__(self):
        super(SavePy, self).__init__("savepy", gdb.COMMAND_OBSCURE)

    def invoke(self, args, from_tty):
        filename, var = args.split()
        data = data_extractor.extract_var(var)
        with open(filename, "wb") as file:
            pickle.dump(data, file)


class Save(gdb.Command):
    def __init__(self):
        super(Save, self).__init__("save", gdb.COMMAND_OBSCURE)

    def invoke(self, args, from_tty):
        filename, var = args.split()
        data = data_extractor.extract_var(var)
        data.tofile(filename)


SaveMat()
SavePy()
Save()


import re
from typing import Tuple, Optional

import gdb  # pylint: disable=E0401
import gdb.types  # pylint: disable=E0401
import numpy as np

from .type_handler import TypeHandler, ScalarTypeHandler

COMPLEX_REGEX = re.compile("(\\S*) \\+ (\\S*) \\* I")


class StdVector(TypeHandler):
    @staticmethod
    def can_handle(gdb_type: gdb.Type) -> bool:
        return str(gdb_type).startswith("std::vector")

    def shape(self, gdb_value: gdb.Value) -> Tuple[Optional[int], ...]:
        size = int(gdb_value["_M_impl"]["_M_finish"] - gdb_value["_M_impl"]["_M_start"])
        return (size,)

    def contained_type(self, gdb_value: gdb.Value) -> gdb.Type:
        return gdb_value.type.template_argument(0)

    def extract(self, gdb_value: gdb.Value, index: Tuple[int, ...]):
        return (gdb_value["_M_impl"]["_M_start"] + index[0]).dereference()


class StdArray(TypeHandler):
    @staticmethod
    def can_handle(gdb_type: gdb.Type) -> bool:
        return str(gdb_type).startswith("std::array")

    def shape(self, gdb_value: gdb.Value) -> Tuple[Optional[int], ...]:
        size = int(gdb_value.type.template_argument(1))
        return (size,)

    def contained_type(self, gdb_value: gdb.Value) -> Optional[gdb.Type]:
        return gdb_value.type.template_argument(0)

    def extract(self, gdb_value: gdb.Value, index: Tuple[int, ...]):
        return gdb_value["_M_elems"][index[0]]


class Pointer(TypeHandler):
    @staticmethod
    def can_handle(gdb_type: gdb.Type) -> bool:
        return gdb_type.code == gdb.TYPE_CODE_PTR

    def shape(self, gdb_value: gdb.Value) -> Tuple[Optional[int], ...]:
        return (None,)

    def contained_type(self, gdb_value: gdb.Value) -> Optional[gdb.Type]:
        return gdb_value.type.target()

    def extract(self, gdb_value: gdb.Value, index: Tuple[int, ...]):
        return gdb_value[index[0]]


class Array(TypeHandler):
    @staticmethod
    def can_handle(gdb_type: gdb.Type) -> bool:
        return gdb_type.code == gdb.TYPE_CODE_ARRAY

    def shape(self, gdb_value: gdb.Value) -> Tuple[Optional[int], ...]:
        size = gdb_value.type.range()[1]
        return (size, )

    def contained_type(self, gdb_value: gdb.Value) -> Optional[gdb.Type]:
        return gdb_value.type.target()

    def extract(self, gdb_value: gdb.Value, index: Tuple[int, ...]):
        return gdb_value[index[0]]


class Double(ScalarTypeHandler):
    @staticmethod
    def can_handle(gdb_type: gdb.Type) -> bool:
        return str(gdb_type) == "double"

    def extract(self, gdb_value: gdb.Value, index: Tuple[int, ...]):
        return np.float64(gdb_value)


class Float(ScalarTypeHandler):
    @staticmethod
    def can_handle(gdb_type: gdb.Type) -> bool:
        return str(gdb_type) == "float"

    def extract(self, gdb_value: gdb.Value, index: Tuple[int, ...]):
        return np.float32(gdb_value)


def extract_complex(gdb_value) -> complex:
    complex_str = str(gdb_value["_M_value"])
    complex_match = COMPLEX_REGEX.search(complex_str)
    real, imag = complex_match.group(1), complex_match.group(2)
    return float(real) + 1j * float(imag)


class StdComplexDouble(ScalarTypeHandler):
    @staticmethod
    def can_handle(gdb_type: gdb.Type) -> bool:
        return str(gdb_type) == "std::complex<double>"

    def extract(self, gdb_value: gdb.Value, index: Tuple[int, ...]):
        return np.complex128(extract_complex(gdb_value))


class StdComplexFloat(ScalarTypeHandler):
    @staticmethod
    def can_handle(gdb_type: gdb.Type) -> bool:
        return str(gdb_type) == "std::complex<float>"

    def extract(self, gdb_value: gdb.Value, index: Tuple[int, ...]):
        return np.complex64(extract_complex(gdb_value))


class Integral(ScalarTypeHandler):
    @staticmethod
    def can_handle(gdb_type: gdb.Type) -> bool:
        return str(gdb_type) in (
            "char", "short", "int", "long", "long long",
            "unsigned char", "unsigned short", "unsigned int", "unsigned long", "unsigned long long"
        )

    def shape(self, gdb_value: gdb.Value) -> Tuple[Optional[int], ...]:
        return ()

    def contained_type(self, gdb_value: gdb.Value) -> Optional[gdb.Type]:
        dtype = str(gdb.types.get_basic_type(gdb_value.type))
        prefix = "u" if "unsigned" in dtype else "i"
        if "char" in dtype:
            size = "1"
        elif "short" in dtype:
            size = "2"
        elif "int" in dtype:
            size = "4"
        else:
            size = "8"

        self.np_dtype = np.dtype(prefix + size)
        return None

    def extract(self, gdb_value: gdb.Value, index: Tuple[int, ...]):
        return self.np_dtype.type(gdb_value)



from abc import ABC, abstractmethod
from typing import List, Tuple, Optional

import gdb.types  # pylint: disable=E0401
import gdb  # pylint: disable=E0401
import numpy as np

from .type_set import TypeSet
from . import util


class TypeHandler(ABC):
    @staticmethod
    @abstractmethod
    def can_handle(gdb_type: gdb.Type) -> bool:
        """
        Returns whether the handler is able to extract values from a given type

        Parameters:
        gdb_type (gdb.Type): Type to be handled

        Returns:
        bool: Can the type be handled?
        """
        pass

    def __init__(self, type_set: TypeSet):
        self.type_set = type_set

    @abstractmethod
    def shape(self, gdb_value: gdb.Value) -> Tuple[Optional[int], ...]:
        """
        Gets the shape of a container

        Parameters:
        gdb_value (gdb.Value): The container

        Returns:
        Tuple[Optional[int]: The shape of the container. Any dimensions with
                             unbounded size are represented by None
        """
        pass

    @abstractmethod
    def contained_type(self, gdb_value: gdb.Value) -> Optional[gdb.Type]:
        """
        Gets the type of the elements of the container

        Parameters:
        gdb_value (gdb.Value): The container

        Returns:
        Optional[gdb.Type]: The element type. If the value is a scalar, and as
                            such does not hold elements, None is returned
        """
        pass

    @abstractmethod
    def extract(self, gdb_value: gdb.Value, index: Tuple[int, ...]):
        """
        Extracts a value from a GDB value

        Parameters:
        gdb_value (gdb.Value): GDB value
        index (Tuple[int, ...]): The multi-dimensional index of the value to
                                 extract

        Returns:
        value: The extracted value. Either a GDB value if gdb_value was a
               container, or a Numpy dtype if it was a scalar type
        """
        pass

    def extract_all(self, gdb_value: gdb.Value, slices: List[slice]):
        shape = self.shape(gdb_value)
        contained_type = self.contained_type(gdb_value)
        n_dims = len(shape)
        scalar_type = (contained_type == None)

        if scalar_type:
            return self.extract(gdb_value, None)

        basic_contained_type = gdb.types.get_basic_type(contained_type)
        contained_handler = self.type_set.get_handler(basic_contained_type)
        current_slices = slices[:n_dims]
        contained_slices = slices[n_dims:]

        for _ in range(len(shape) - len(current_slices)):
            current_slices.append(slice(None, None, None))

        def gen_output(slc, shp, index):
            if not shp:
                contained_gdb_value = self.extract(gdb_value, index)
                return contained_handler.extract_all(contained_gdb_value, contained_slices)
            else:
                out = []
                for i in util.indices_1d(slc[0], shp[0]):
                    out.append(gen_output(slc[1:], shp[1:], (*index, i)))

                return out

        out = gen_output(current_slices, shape, ())
        return out


class ScalarTypeHandler(TypeHandler):
    def shape(self, gdb_value: gdb.Value) -> Tuple[Optional[int], ...]:
        return ()

    def contained_type(self, gdb_value: gdb.Value) -> Optional[gdb.Type]:
        return None



from .type_set import TypeSet
from . import std_types

default = TypeSet()
default.register(std_types.StdVector)
default.register(std_types.StdArray)
default.register(std_types.Pointer)
default.register(std_types.Array)
default.register(std_types.Double)
default.register(std_types.Float)
default.register(std_types.StdComplexDouble)
default.register(std_types.StdComplexFloat)
default.register(std_types.Integral)


import re
from typing import List

import gdb  # pylint: disable=E0401
import gdb.types  # pylint: disable=E0401
import numpy as np

from .default import default
from .type_set import TypeSet
from . import util


class SliceSyntaxError(Exception):
    pass


class VariableError(Exception):
    pass


def parse_subslice(subslice: str) -> slice:
    slice_components = subslice.split(":")

    try:
        s = [int(gdb.parse_and_eval(x)) if x else None for x in slice_components]
    except (ValueError, gdb.error):
        raise SliceSyntaxError(f"Invalid slice component: {subslice}")

    if len(s) == 1:
        return slice(s[0], s[0] + 1)
    if len(s) == 2:
        return slice(s[0], s[1])
    elif len(s) == 3:
        return slice(s[0], s[1], s[2])
    else:
        raise SliceSyntaxError(f"Invalid slice component: {subslice}")


def parse_slice(full_slice: str) -> List[slice]:
    slice_dims = full_slice.split(",")
    return [parse_subslice(s) for s in slice_dims]


def parse_var(var: str):
    var_base, slice_str = util.split_variable_and_slice(var)
    slices = parse_slice(slice_str) if slice_str else []
    return var_base, slices


def extract_var(var: str, type_set: TypeSet = default) -> np.ndarray:
    base_var, var_slice = parse_var(var)

    try:
        gdb_data = gdb.parse_and_eval(base_var)
    except gdb.error:
        raise VariableError(f"Invalid variable: {var}")

    gdb_type = gdb.types.get_basic_type(gdb_data.type)
    type_handler = type_set.get_handler(gdb_type)

    out = type_handler.extract_all(gdb_data, var_slice)
    out_np = np.squeeze(np.array(out))
    return out_np



import gdb # pylint: disable=E0401
import matplotlib.pyplot as plt
import mpl_toolkits.mplot3d.axes3d as p3
import numpy as np

from . import data_extractor


class PlottingError(Exception):
    pass


class Legend:
    def __init__(self):
        self.legend = []

    def add(self, name: str):
        self.legend.append(name)

    def apply(self):
        plt.legend(self.legend)


def plot_1d(args, plot_function):
    legend = Legend()

    for arg in args.split():
        data = data_extractor.extract_var(arg)

        if data.ndim == 2 and not np.iscomplexobj(data):
            for i, row in enumerate(data):
                plot_function(row)
                legend.add(f"{arg}[{i}]")
        elif data.ndim == 1:
            if np.iscomplexobj(data):
                plot_function(np.real(data))
                plot_function(np.imag(data))
                plot_function(np.abs(data))
                legend.add(f"real({arg})")
                legend.add(f"imag({arg})")
                legend.add(f"abs({arg})")
            else:
                plot_function(data)
                legend.add(arg)
        else:
            raise PlottingError(f"Unsuitable for plotting: {arg}")
    
    legend.apply()
    plt.grid()
    plt.show()


class Plot(gdb.Command):
    def __init__(self):
        super(Plot, self).__init__("plot", gdb.COMMAND_OBSCURE)

    def invoke(self, args, from_tty):
        plot_1d(args, plt.plot)


class Scatter(gdb.Command):
    def __init__(self):
        super(Scatter, self).__init__("scatter", gdb.COMMAND_OBSCURE)

    def invoke(self, args, from_tty):
        legend = Legend()
        temp = []

        for arg in args.split():
            data = data_extractor.extract_var(arg)

            if data.ndim == 2 and not np.iscomplexobj(data) and 2 in data.shape:
                if data.shape[1] == 2:
                    plt.scatter(data[:,0], data[:,1])
                else:
                    plt.scatter(data[0], data[1])

                legend.add(arg)
            elif data.ndim == 1:
                if np.iscomplexobj(data):
                    plt.scatter(np.real(data), np.imag(data))
                    legend.add(arg)
                else:
                    temp.append(data)
            else:
                raise PlottingError(f"Unsuitable for plotting: {arg}")

        if len(temp):
            if len(temp) == 2:
                plt.scatter(temp[0], temp[1])
                legend.add("Data")
            else:
                raise PlottingError(f"Incorrect number of arguments")
        
        legend.apply()
        plt.grid()
        plt.show()


class Plot3D(gdb.Command):
    def __init__(self):
        super(Plot3D, self).__init__("plot3d", gdb.COMMAND_OBSCURE)

    def invoke(self, args, from_tty):
        z = data_extractor.extract_var(args)
        if z.ndim != 2:
            raise PlottingError(f"Unsuitable for plotting: {args}")

        x = np.arange(z.shape[1])
        y = np.arange(z.shape[0])
        xm, ym = np.meshgrid(x, y)

        fig = plt.figure()
        ax = p3.Axes3D(fig)
        ax.plot_surface(xm, ym, z)
        plt.show()


class Scatter3D(gdb.Command):
    def __init__(self):
        super(Scatter3D, self).__init__("scatter3d", gdb.COMMAND_OBSCURE)

    def invoke(self, args, from_tty):
        legend = Legend()
        temp = []

        fig = plt.figure()
        ax = p3.Axes3D(fig)

        for arg in args.split():
            data = data_extractor.extract_var(arg)

            if data.ndim == 2 and not np.iscomplexobj(data) and 3 in data.shape:
                if data.shape[1] == 3:
                    ax.scatter(data[:,0], data[:,1], data[:,2])
                else:
                    ax.scatter(data[0], data[1], data[2])

                legend.add(arg)
            elif data.ndim == 1:
                temp.append(data)  
            else:
                raise PlottingError(f"Unsuitable for plotting: {arg}")

        if len(temp):
            if len(temp) == 3:
                ax.scatter(temp[0], temp[1])
                legend.add("Data")
            else:
                raise PlottingError(f"Incorrect number of arguments")
        
        legend.apply()
        plt.grid()
        plt.show()


class Hist(gdb.Command):
    def __init__(self):
        super(Hist, self).__init__("hist", gdb.COMMAND_OBSCURE)

    def invoke(self, args, from_tty):
        hist = lambda x: plt.hist(x, bins="auto")
        plot_1d(args, hist)


class FFT(gdb.Command):
    def __init__(self):
        super(FFT, self).__init__("fft", gdb.COMMAND_OBSCURE)

    def invoke(self, args, from_tty):
        legend = Legend()
        fft_db = lambda x: 20*np.log10(np.abs(np.fft.fft(x)))

        for arg in args.split():
            data = data_extractor.extract_var(arg)

            if data.ndim == 2 and not np.iscomplexobj(data):
                for i, row in enumerate(data):
                    plt.plot(fft_db(row))
                    legend.add(f"{arg}[{i}]")
            elif data.ndim == 1:
                plt.plot(fft_db(data))
                legend.add(arg)
            else:
                raise PlottingError(f"Unsuitable for plotting: {arg}")

        legend.apply()
        plt.grid()
        plt.ylabel("PSD (dB)")
        plt.show()


Plot()
Scatter()
Plot3D()
Scatter3D()
Hist()
FFT()


import re
from typing import List, Tuple, Optional


def indices_1d(s: slice, shape: int):
    if shape is None:
        start = 0 if s.start is None else s.start
        stop = 0 if s.stop is None else s.stop
        step = 1 if s.step is None else s.step
    else:
        start, stop, step = s.indices(shape)

    in_range = lambda a, b, c: a < b if c > 0 else a > b
    i = start
    while in_range(i, stop, step):
        yield i
        i += step


def indices(slices: List[slice], shape: Tuple):
    if len(slices) != len(shape):
        for i in range(len(shape) - len(slices)):
            slices.append(slice(None, None, None))


    if len(slices) == 1:
        for i in indices_1d(slices[0], shape[0]):
            yield (i,)
    else:
        for i in indices_1d(slices[0], shape[0]):
            for j in indices(slices[1:], shape[1:]):
                yield (i, *j)


def split_variable_and_slice(var: str) -> Tuple[str, Optional[str]]:
    lbracket = var.rfind("[")
    rbracket = var.rfind("]")

    if lbracket == -1 or rbracket == -1:
        return var, None
    else:
        return var[:lbracket], var[lbracket+1:rbracket]


def strip_non_alphanumeric(s: str) -> str:
    return re.sub("\\W+", "", s)


import setuptools

with open('README.md') as readme:
    long_description = readme.read()

setuptools.setup(
    name='gdbplotlib',
    version='0.1.0',
    author='George Cholerton',
    author_email='gcholerton@gmail.com',
    url='https://github.com/X-Neon/gdbplotlib',
    packages=setuptools.find_packages(),
    install_requires=['numpy'],
    python_requires='>=3',
    description='Plotting and exporting of variables from GDB',
    long_description=long_description,
    long_description_content_type='text/markdown',
    keywords=['gdb', 'debug'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Topic :: Scientific/Engineering :: Visualization',
        'Topic :: Software Development :: Debuggers'
    ]
)


class UnkownTypeError(Exception):
    pass


class TypeSet:
    def __init__(self):
        self.handlers = []

    def register(self, type_handler):
        self.handlers.append(type_handler)

    def get_handler(self, gdb_type):
        for handler in self.handlers:
            if handler.can_handle(gdb_type):
                return handler(self)

        raise UnkownTypeError(f"Cannot handle type: {str(gdb_type)}")
